@@ -26,7 +26,7 @@
  * It's recommended to not change any alignment or device properties. All is
  * initialized by default by fdisk_assign_device().
  *
- * Note that terminology used by libfdisk is: 
+ * Note that terminology used by libfdisk is:
  *   - device properties: I/O limits (topology), geometry, sector size, ...
  *   - alignment: first, last LBA, grain, ...
  *
@@ -110,17 +110,17 @@ fdisk_sector_t fdisk_align_lba(struct fdisk_context *cxt, fdisk_sector_t lba, in
 				res += sects_in_phy;
 		}
 	}
-
+/*
 	if (lba != res)
-		DBG(CXT, ul_debugobj(cxt, "LBA %ju -aligned-%s-> %ju [grain=%lus]",
+		DBG(CXT, ul_debugobj(cxt, "LBA %12ju aligned-%s %12ju [grain=%lus]",
 				(uintmax_t) lba,
-				direction == FDISK_ALIGN_UP ? "up" :
+				direction == FDISK_ALIGN_UP ? "up  " :
 				direction == FDISK_ALIGN_DOWN ? "down" : "near",
 				(uintmax_t) res,
 				cxt->grain / cxt->sector_size));
 	else
-		DBG(CXT, ul_debugobj(cxt, "LBA %ju -unchanged-", (uintmax_t)lba));
-
+		DBG(CXT, ul_debugobj(cxt, "LBA %12ju already aligned", (uintmax_t)lba));
+*/
 	return res;
 }
 
@@ -140,12 +140,14 @@ fdisk_sector_t fdisk_align_lba_in_range(struct fdisk_context *cxt,
 {
 	fdisk_sector_t res;
 
-	start = fdisk_align_lba(cxt, start, FDISK_ALIGN_UP);
-	stop = fdisk_align_lba(cxt, stop, FDISK_ALIGN_DOWN);
+	/*DBG(CXT, ul_debugobj(cxt, "LBA: align in range <%ju..%ju>", (uintmax_t) start, (uintmax_t) stop));*/
 
-	if (lba > start && lba < stop
-	    && (lba - start) < (cxt->grain / cxt->sector_size)) {
+	if (start + (cxt->grain / cxt->sector_size) <= stop) {
+		start = fdisk_align_lba(cxt, start, FDISK_ALIGN_UP);
+		stop = fdisk_align_lba(cxt, stop, FDISK_ALIGN_DOWN);
+	}
 
+	if (start + (cxt->grain / cxt->sector_size) > stop) {
 		DBG(CXT, ul_debugobj(cxt, "LBA: area smaller than grain, don't align"));
 		res = lba;
 		goto done;
@@ -159,9 +161,8 @@ fdisk_sector_t fdisk_align_lba_in_range(struct fdisk_context *cxt,
 		res = stop;
 	else
 		res = lba;
-
 done:
-	DBG(CXT, ul_debugobj(cxt, "LBA %ju range:<%ju..%ju>, result: %ju",
+	DBG(CXT, ul_debugobj(cxt, "%ju in range <%ju..%ju> aligned to %ju",
 				(uintmax_t) lba,
 				(uintmax_t) start,
 				(uintmax_t) stop,
@@ -313,6 +314,37 @@ int fdisk_save_user_sector_size(struct fdisk_context *cxt,
 	return 0;
 }
 
+/**
+ * fdisk_save_user_grain:
+ * @cxt: context
+ * @grain: size in bytes (>= 512, multiple of 512)
+ *
+ * Save user define grain size. The size is used to align partitions.
+ *
+ * The default is 1MiB (or optimal I/O size if greater than 1MiB). It's strongly
+ * recommended to use the default.
+ *
+ * The smallest possible granularity for partitioning is physical sector size
+ * (or minimal I/O size; the bigger number win). If the user's @grain size is
+ * too small then the smallest possible granularity is used. It means
+ * fdisk_save_user_grain(cxt, 512) forces libfdisk to use grain as small as
+ * possible.
+ *
+ * The setting is applied by fdisk_assign_device() or
+ * fdisk_reset_device_properties().
+ *
+ * Returns: <0 on error, 0 on success.
+ */
+int fdisk_save_user_grain(struct fdisk_context *cxt, unsigned long grain)
+{
+	if (!cxt || grain % 512)
+		return -EINVAL;
+
+	DBG(CXT, ul_debugobj(cxt, "user grain size: %lu", grain));
+	cxt->user_grain = grain;
+	return 0;
+}
+
 /**
  * fdisk_has_user_device_properties:
  * @cxt: context
@@ -322,6 +354,7 @@ int fdisk_save_user_sector_size(struct fdisk_context *cxt,
 int fdisk_has_user_device_properties(struct fdisk_context *cxt)
 {
 	return (cxt->user_pyh_sector || cxt->user_log_sector ||
+		cxt->user_grain ||
 		fdisk_has_user_device_geometry(cxt));
 }
 
@@ -363,6 +396,14 @@ int fdisk_apply_user_device_properties(struct fdisk_context *cxt)
 		recount_geometry(cxt);
 
 	fdisk_reset_alignment(cxt);
+
+	if (cxt->user_grain) {
+		unsigned long granularity = max(cxt->phy_sector_size, cxt->min_io_size);
+
+		cxt->grain = cxt->user_grain < granularity ? granularity : cxt->user_grain;
+		DBG(CXT, ul_debugobj(cxt, "new grain: %lu", cxt->grain));
+	}
+
 	if (cxt->firstsector_bufsz != cxt->sector_size)
 		fdisk_read_firstsector(cxt);
 
@@ -500,12 +541,16 @@ int fdisk_discover_topology(struct fdisk_context *cxt)
 				/* optimal I/O is optional, default to minimum IO */
 				cxt->io_size = cxt->min_io_size;
 
-			/* ignore optimal I/O if not aligned to phy.sector size */
-			if (cxt->io_size
-			    && cxt->phy_sector_size
-			    && (cxt->io_size % cxt->phy_sector_size) != 0) {
-				DBG(CXT, ul_debugobj(cxt, "ignore misaligned I/O size"));
-				cxt->io_size = cxt->phy_sector_size;
+			if (cxt->io_size && cxt->phy_sector_size) {
+				if (cxt->io_size == 33553920) {
+					/* 33553920 (32 MiB - 512) is always a controller error */
+					DBG(CXT, ul_debugobj(cxt, "ignore bad I/O size 33553920"));
+					cxt->io_size = cxt->phy_sector_size;
+				} else if ((cxt->io_size % cxt->phy_sector_size) != 0) {
+					/* ignore optimal I/O if not aligned to phy.sector size */
+					DBG(CXT, ul_debugobj(cxt, "ignore misaligned I/O size"));
+					cxt->io_size = cxt->phy_sector_size;
+				}
 			}
 
 		}
@@ -622,7 +667,7 @@ int fdisk_apply_label_device_properties(struct fdisk_context *cxt)
 	int rc = 0;
 
 	if (cxt->label && cxt->label->op->reset_alignment) {
-		DBG(CXT, ul_debugobj(cxt, "appling label device properties..."));
+		DBG(CXT, ul_debugobj(cxt, "applying label device properties..."));
 		rc = cxt->label->op->reset_alignment(cxt);
 	}
 	return rc;
@@ -674,43 +719,3 @@ fdisk_sector_t fdisk_cround(struct fdisk_context *cxt, fdisk_sector_t num)
 			(num / fdisk_get_units_per_sector(cxt)) + 1 : num;
 }
 
-/**
- * fdisk_reread_partition_table:
- * @cxt: context
- *
- * Force *kernel* to re-read partition table on block devices.
- *
- * Returns: 0 on success, < 0 in case of error.
- */
-int fdisk_reread_partition_table(struct fdisk_context *cxt)
-{
-	int i;
-	struct stat statbuf;
-
-	assert(cxt);
-	assert(cxt->dev_fd >= 0);
-
-	i = fstat(cxt->dev_fd, &statbuf);
-	if (i == 0 && S_ISBLK(statbuf.st_mode)) {
-		DBG(CXT, ul_debugobj(cxt, "calling re-read ioctl"));
-		sync();
-#ifdef BLKRRPART
-		fdisk_info(cxt, _("Calling ioctl() to re-read partition table."));
-		i = ioctl(cxt->dev_fd, BLKRRPART);
-#else
-		errno = ENOSYS;
-		i = 1;
-#endif
-	}
-
-	if (i) {
-		fdisk_warn(cxt, _("Re-reading the partition table failed."));
-		fdisk_info(cxt,	_(
-			"The kernel still uses the old table. The "
-			"new table will be used at the next reboot "
-			"or after you run partprobe(8) or kpartx(8)."));
-		return -errno;
-	}
-
-	return 0;
-}
